# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke, Florian Roth (Nextron Systems), juju4

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import json
import os
import re
import sys

from sigma.backends.base import SingleTextQueryBackend
from sigma.backends.exceptions import NotSupportedError
from sigma.parser.condition import ConditionOR, SigmaAggregationParser

# Sumo specifics
# https://help.sumologic.com/05Search/Search-Query-Language
# want _index or _sourceCategory for performance
# try to get most string match on first line for performance
# further sorting can be done with extra parsing
# No regex match, must use 'parse regex' https://help.sumologic.com/05Search/Search-Query-Language/01-Parse-Operators/02-Parse-Variable-Patterns-Using-Regex
# For some strings like Windows ProcessCmdline or LogonProcess, it might be good to force case lower and upper as Windows is inconsistent in logs


class SumoLogicBackend(SingleTextQueryBackend):
    """Converts Sigma rule into SumoLogic query. Contributed by SOC Prime. https://socprime.com"""
    identifier = "sumologic"
    active = True
    config_required = False
    default_config = ["sysmon", "sumologic"]

    index_field = "_sourceCategory"
    reClear = None
    andToken = " AND "
    orToken = " OR "
    notToken = "!"
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = ", "
    valueExpression = "\"%s\""
    nullExpression = "isEmpty(%s)"
    notNullExpression = "!isEmpty(%s)"
    mapExpression = "%s=%s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s IN %s"
    interval = None
    logname = None

    def generateAggregation(self, agg):
        # lnx_shell_priv_esc_prep.yml
        # print("DEBUG generateAggregation(): %s, %s, %s, %s" % (agg.aggfunc_notrans, agg.aggfield, agg.groupfield, agg.cond_op))
        if agg.groupfield == 'host':
            agg.groupfield = 'hostname'
        if agg.aggfunc_notrans == 'count() by':
            agg.aggfunc_notrans = 'count by'
        if agg.aggfunc == SigmaAggregationParser.AGGFUNC_NEAR:
            raise NotImplementedError("The 'near' aggregation operator is not yet implemented for this backend")
        if self.keypresent:
            if not agg.groupfield:
                if agg.aggfield:
                    agg.aggfunc_notrans = "count_distinct"
                    return " \n| %s(%s) \n| where _count_distinct %s %s" % (
                        agg.aggfunc_notrans, agg.aggfield, agg.cond_op, agg.condition)
                else:
                    return "  \n| %s | where _count %s %s" % (
                    agg.aggfunc_notrans, agg.cond_op, agg.condition)
            elif agg.groupfield:
                if agg.aggfield:
                    agg.aggfunc_notrans = "count_distinct"
                    return " \n| %s(%s) by %s \n| where _count_distinct %s %s" % (
                        agg.aggfunc_notrans, agg.aggfield, agg.groupfield, agg.cond_op, agg.condition)
                else:
                    return " \n| %s by %s \n| where _count %s %s" % (
                        agg.aggfunc_notrans, agg.groupfield, agg.cond_op, agg.condition)
            else:
                return " \n| %s | where _count %s %s" % (agg.aggfunc_notrans, agg.cond_op, agg.condition)
        else:
            if not agg.groupfield:
                if agg.aggfield:
                    agg.aggfunc_notrans = "count_distinct"
                    return " \n| parse \"[%s=*]\" as searched nodrop\n| %s(searched) \n| where _count_distinct %s %s" % (
                        agg.aggfield, agg.aggfunc_notrans, agg.cond_op, agg.condition)
                else:
                    return " \n| %s | where _count %s %s" % (
                    agg.aggfunc_notrans, agg.cond_op, agg.condition)
            elif agg.groupfield:
                if agg.aggfield:
                    agg.aggfunc_notrans = "count_distinct"
                    return " \n| parse \"[%s=*]\" as searched nodrop\n| parse \"[%s=*]\" as grpd nodrop\n| %s(searched) by grpd \n| where _count_distinct %s %s" % (
                        agg.aggfield, agg.groupfield, agg.aggfunc_notrans, agg.cond_op, agg.condition)
                else:
                    return " \n| parse \"[%s=*]\" as grpd nodrop\n| %s by grpd \n| where _count %s %s" % (
                        agg.groupfield, agg.aggfunc_notrans, agg.cond_op, agg.condition)
            else:
                return " \n| %s | where _count %s %s" % (agg.aggfunc_notrans, agg.cond_op, agg.condition)

    def generateBefore(self, parsed):
        # not required but makes query faster, especially if no FER or _index/_sourceCategory
        if self.logname:
            return "%s " % self.logname
        return ""

    def generate(self, sigmaparser):
        try:
            self.product = sigmaparser.parsedyaml['logsource']['product']   # OS or Software
        except KeyError:
            self.product = None
        try:
            self.service = sigmaparser.parsedyaml['logsource']['service']   # Channel
        except KeyError:
            self.service = None
        try:
            self.category = sigmaparser.parsedyaml['logsource']['category']   # Channel
        except KeyError:
            self.category = None
        # FIXME! don't get backend config mapping
        self.indices = sigmaparser.get_logsource().index
        if len(self.indices) == 0:
            self.indices = None
        try:
            self.interval = sigmaparser.parsedyaml['detection']['timeframe']
        except:
            pass

        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed)
            # FIXME! exclude if expression is regexp but anyway, not directly supported.
            #   Not doing if aggregation ('| count') or key ('=')
            if not (query.startswith('"') and query.endswith('"')) and not (query.startswith('(') and query.endswith(')')) and not ('|' in query) and not ('=' in query):
                query = '"%s"' % query
            before = self.generateBefore(parsed)
            after = self.generateAfter(parsed)

            result = ""
            if before is not None:
                result = before
            if query is not None:
                result += query
            if after is not None:
                result += after

            # adding parenthesis here in case 2 rules are aggregated together - ex: win_possible_applocker_bypass
            # but does not work if count, where or other piped statements...
            if '|' in result:
                return result
            else:
                return result

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO/FIXME! depending on deployment configuration, existing FER must be populate here (or backend config?)
        aFL = ["_sourceCategory", "_view", "_sourceName"]
        if self.sigmaconfig.config.get("afl_fields"):
            self.keypresent = True
            aFL.extend(self.sigmaconfig.config.get("afl_fields"))
        else:
            self.keypresent = False
        for item in self.sigmaconfig.fieldmappings.values():
            if item.target_type is list:
                aFL.extend(item.target)
            else:
                aFL.append(item.target)
        self.allowedFieldsList = list(set(aFL))

    # Skip logsource value from sigma document for separate path.
    # def generateCleanValueNodeLogsource(self, value):
    #    return self.valueExpression % (self.cleanValue(str(value)))

    # Clearing values from special characters.
    # Sumologic: only removing '*' (in quotes, is literal. without, is wildcard) and '"'

    def cleanNode(self, node, key=None):
        if "*" in node and key and not re.search("[\s]", node):
            return node
        elif "*" in node and not key:
            return [x for x in node.split("*") if x]
        return node

    # Clearing values from special characters.
    def generateMapItemNode(self, node):
        key, value = node
        if key in self.allowedFieldsList:
            if key in ["_sourceCategory", "_sourceName"]:
                value = "*%s*" % value.lower()
                return self.mapExpression % (key, value)
            elif not self.mapListsSpecialHandling and type(value) in (
                    str, int, list) or self.mapListsSpecialHandling and type(value) in (str, int):
                if key in ("LogName", "source"):
                    self.logname = value
                # need cleanValue if sigma entry with single quote
                return self.mapExpression % (key, self.cleanValue(value, key))
            elif type(value) is list:
                return self.generateMapItemListNode(key, value)
            elif value is None:
                return self.nullExpression % (key, )
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))
        else:
            if not self.mapListsSpecialHandling and type(value) in (
                    str, int, list) or self.mapListsSpecialHandling and type(value) in (str, int):
                if type(value) is str:
                    new_value = list()
                    value = self.cleanNode(value)
                    if type(value) == list:
                        new_value.append(self.andToken.join([self.cleanValue(val) for val in value]))
                    else:
                        new_value.append(value)
                    if len(new_value) == 1:
                        if self.generateANDNode(new_value):
                            return self.generateANDNode(new_value)
                        else:
                            # if after cleaning node, it is empty but there is AND statement... make it true.
                            return "true"
                    else:
                        return self.generateORNode(new_value)
                else:
                    return self.generateValueNode(value)
            elif type(value) is list:
                new_value = list()
                for item in value:
                    item = self.cleanNode(item)
                    if type(item) is list and len(item) == 1:
                        new_value.append(item[0])
                    elif type(item) is list:
                        new_value.append(self.andToken.join([self.cleanValue(val) for val in item]))
                    else:
                        new_value.append(item)
                return self.generateORNode(new_value)
            elif value is None:
                return self.nullExpression % (key, )
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))

    # from mixins.py
    # input in simple quotes are not passing through this function. ex: rules/windows/sysmon/sysmon_vul_java_remote_debugging.yml, rules/apt/apt_sofacy_zebrocy.yml
    #   => OK only if field entry with list, not string
    #   => generateNode: call cleanValue
    def cleanValue(self, val, key=''):
        if isinstance(val, str):
            val = re.sub("[^\\\"](\")", "\\\"", val)
            if re.search("[\W\s]", val):# and not val.startswith('"') and not val.endswith('"'):  # or "\\" in node in [] or "/" in node:
                return self.valueExpression % val
        return val

    # for keywords values with space
    def generateValueNode(self, node, key=''):
        cV = self.cleanNode(str(node), key)
        if type(node) is int:
            return cV
        if type(cV) is list:
            return "(%s)" % "AND".join([self.cleanValue(item) for item in cV])
        if 'AND' in node and cV:
            return "(" + cV + ")"
        elif isinstance(node, str) and node.startswith('"') and node.endswith('"'):
            return cV
        else:
            return self.cleanValue(cV)

    def generateMapItemListNode(self, key, value):
        itemslist = list()
        for item in value:
            if key in self.allowedFieldsList:
                itemslist.append('%s = %s' % (key, self.generateValueNode(item, key)))
            else:
                itemslist.append('%s' % (self.generateValueNode(item)))
        return "(" + " OR ".join(itemslist) + ")"

    # generateORNode algorithm for SumoLogicBackend class.
    def generateORNode(self, node):
        if type(node) == ConditionOR and all(isinstance(item, str) for item in node):
            new_value = list()
            for value in node:
                value = self.cleanNode(value)
                if type(value) is list:
                    new_value.append(self.andToken.join([self.valueExpression % val for val in value]))
                else:
                    new_value.append(value)
            return "(" + self.orToken.join([self.generateNode(val) for val in new_value]) + ")"
        return "(" + self.orToken.join([self.generateNode(val) for val in node]) + ")"


class SumoLogicCSE(SumoLogicBackend):
    """Converts Sigma rule into SumoLogic CSE query. Contributed by SOC Prime. https://socprime.com"""
    identifier = "sumologic-cse"
    active = True
    config_required = False
    default_config = ["sysmon"]

    index_field = "metdata_product"
    reClear = None
    #reEscape = re.compile('[\\\\"]')
    andToken = " and "
    orToken = " or "
    notToken = "!"
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = ", "
    valueExpression = "\"%s\""
    nullExpression = "isEmpty(%s)"
    notNullExpression = "!isEmpty(%s)"
    mapExpression = "%s=%s"
    mapListsSpecialHandling = True
    mapListValueExpression = "%s IN %s"
    interval = None
    logname = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.allowedFieldsList.extend(["metdata_product", "metdata_vendor"])

    def cleanValue(self, val, key=''):
        if key == 'metadata_deviceEventId' or isinstance(val, int) or val.isdigit():
            return val
        return self.valueExpression % val

    def cleanNode(self, node, key=None):
        return node

    # Clearing values from special characters.
    def generateMapItemNode(self, node):
        key, value = node
        if key:
            if not self.mapListsSpecialHandling and type(value) in (
                    str, int, list) or self.mapListsSpecialHandling and type(value) in (str, int):
                if key in ("LogName", "source"):
                    self.logname = value
                # need cleanValue if sigma entry with single quote
                return self.mapExpression % (key, self.cleanValue(value, key))
            elif type(value) is list:
                return self.generateMapItemListNode(key, value)
            elif value is None:
                return self.nullExpression % (key,)
            else:
                raise TypeError("Backend does not support map values of type " + str(type(value)))
        raise TypeError("Backend does not support query without key.")

    def generateMapItemListNode(self, key, value):
        if len(value) == 1:
            return self.mapExpression % (key, value[0])
        return "%s IN (%s)" % (key, ", ".join([self.cleanValue(item, key) for item in value]))


class SumoLogicCSERule(SumoLogicCSE):
    """Converts Sigma rule into SumoLogic CSE query"""
    identifier = "sumologic-cse-rule"
    active = True

    def __init__(self, *args, **kwargs):
        """Initialize field mappings"""
        super().__init__(*args, **kwargs)
        self.techniques = self._load_mitre_file("techniques")
        self.allowedCategories = ["Threat Intelligence", "Initial Access", "Execution", "Persistence", "Privilege Escalation",
                                  "Defense Evasion", "Credential Access", "Discovery", "Lateral Movement", "Collection",
                                  "Command and Control", "Exfiltration", "Impact"]
        self.defaultCategory = "Unknown/Other"
        self.results = []

    def find_technique(self, key_ids):
        for key_id in set(key_ids):
            if not key_id:
                continue
            for technique in self.techniques:
                if key_id == technique.get("technique_id", ""):
                    yield technique

    def _load_mitre_file(self, mitre_type):
        try:
            backend_dir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "config", "mitre"))
            path = os.path.join(backend_dir, "{}.json".format(mitre_type))
            with open(path) as config_file:
                config = json.load(config_file)
                return config
        except (IOError, OSError) as e:
            print("Failed to open {} configuration file '%s': %s".format(path, str(e)), file=sys.stderr)
            return []
        except json.JSONDecodeError as e:
            print("Failed to parse {} configuration file '%s' as valid YAML: %s" % (path, str(e)), file=sys.stderr)
            return []

    def skip_tactics_or_techniques(self, src_technics, src_tactics):
        tactics = set()
        technics = set()

        local_storage_techniques = {item["technique_id"]: item for item in self.find_technique(src_technics)}

        for key_id in src_technics:
            src_tactic = local_storage_techniques.get(key_id, {}).get("tactic")
            if not src_tactic:
                continue
            src_tactic = set(src_tactic)

            for item in src_tactics:
                if item in src_tactic:
                    technics.add(key_id)
                    tactics.add(item)

        return sorted(tactics), sorted(technics)

    def parse_severity(self, old_severity):
        if old_severity.lower() == "critical":
            return "high"
        return old_severity

    def get_tactics_and_techniques(self, tags):
        tactics = list()
        technics = list()

        for tag in tags:
            tag = tag.replace("attack.", "")
            if re.match("[t][0-9]{4}", tag, re.IGNORECASE):
                technics.append(tag.title())
            elif re.match("[s][0-9]{4}", tag, re.IGNORECASE):
                continue
            else:
                if "_" in tag:
                    tag = tag.replace("_", " ")
                tag = tag.title()
                tactics.append(tag)

        return tactics, technics

    def map_risk_score(self, level):
        if level == "critical":
            return 5
        elif level == "high":
            return 4
        elif level == "medium":
            return 3
        elif level == "low":
            return 2
        return 1

    def create_rule(self, config):
        tags = config.get("tags", [])

        tactics, technics = self.get_tactics_and_techniques(tags)
        tactics, technics = self.skip_tactics_or_techniques(technics, tactics)
        tactics = list(map(lambda s: s.replace(" ", ""), tactics))
        score = self.map_risk_score(config.get("level", "medium"))
        rule = {
            "name": "{} by {}".format(config.get("title"), config.get('author')),
            "description": "{} {}".format(config.get("description"), "Technique: {}.".format(",".join(technics))),
            "enabled": True,
            "expression": """{}""".format(config.get("translation", "")),
            "assetField": "device_hostname",
            "score": score,
            "stream": "record"
        }
        if tactics and tactics[0] in self.allowedCategories:
            rule.update({"category": tactics[0]})
        else:
            rule.update({"category": "Unknown/Other"})
        self.results.append(rule)
        #return json.dumps(rule, indent=4, sort_keys=False)

    def generate(self, sigmaparser):
        translation = super().generate(sigmaparser)
        if translation:
            configs = sigmaparser.parsedyaml
            configs.update({"translation": translation})
            rule = self.create_rule(configs)
            return rule
        else:
            raise NotSupportedError("No table could be determined from Sigma rule")

    def finalize(self):
        if len(self.results) == 1:
           return json.dumps(self.results[0], indent=4, sort_keys=False)
        elif len(self.results) > 1:
            return json.dumps(self.results, indent=4, sort_keys=False)






# Output backends for sigmac
# Copyright 2018 SOC Prime

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sigma
from .base import SingleTextQueryBackend
from .exceptions import PartialMatchError, FullMatchError
        
class QualysBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Qualys saved search. Contributed by SOC Prime. https://socprime.com"""
    identifier = "qualys"
    active = True
    config_required = False
    default_config = ["sysmon", "qualys"]
    andToken = " and "
    orToken = " or "
    notToken = "not "
    subExpression = "(%s)"
    listExpression = "%s"
    listSeparator = " "
    valueExpression = "%s"
    nullExpression = "%s is null"
    notNullExpression = "not (%s is null)"
    mapExpression = "%s:`%s`"
    mapListsSpecialHandling = True
    PartialMatchFlag = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        fl = []
        for item in self.sigmaconfig.fieldmappings.values():
            if item.target_type == list:
                fl.extend(item.target)
            else:
                fl.append(item.target)
        self.allowedFieldsList = list(set(fl))

    def generateORNode(self, node):
        new_list = []
        for val in node:
            if type(val) == tuple and not(val[0] in self.allowedFieldsList):
                pass
                # self.PartialMatchFlag = True
            else:
                new_list.append(val)

        generated = [self.generateNode(val) for val in new_list]
        filtered = [g for g in generated if g is not None]
        return self.orToken.join(filtered)

    def generateANDNode(self, node):
        new_list = []
        for val in node:
            if type(val) == tuple and not(val[0] in self.allowedFieldsList):
                self.PartialMatchFlag = True
            else:
                new_list.append(val)
        generated = [self.generateNode(val) for val in new_list]
        filtered = [g for g in generated if g is not None]
        return self.andToken.join(filtered)

    def generateMapItemNode(self, node):
        key, value = node
        if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            if key in self.allowedFieldsList:
                return self.mapExpression % (key, self.generateNode(value))
            else:
                return self.generateNode(value)
        elif type(value) == list:
            return self.generateMapItemListNode(key, value)
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateMapItemListNode(self, key, value):
        itemslist = []
        for item in value:
            if key in self.allowedFieldsList:
                itemslist.append('%s:`%s`' % (key, self.generateValueNode(item)))
            else:
                itemslist.append('%s' % (self.generateValueNode(item)))
        return "(" + (" or ".join(itemslist)) + ")"

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        all_keys = set()

        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed)
            if query == "()":
                self.PartialMatchFlag = None

            if self.PartialMatchFlag == True:
                raise PartialMatchError(query)
            elif self.PartialMatchFlag == None:
                raise FullMatchError(query)
            else:
                return query



import re
import requests
import json
import os
from sigma.config.eventdict import event
from fnmatch import fnmatch

from sigma.backends.base import SingleTextQueryBackend
from sigma.backends.exceptions import NotSupportedError
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from sigma.parser.condition import ConditionOR, ConditionAND, NodeSubexpression

from sigma.parser.modifiers.base import SigmaTypeModifier
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier


class CarbonBlackWildcardHandlingMixin:
    """
    Determine field mapping to keyword subfields depending on existence of wildcards in search values. Further,
    provide configurability with backend parameters.
    """

    # options = SingleTextQueryBackend.options + (
    #         ("keyword_field", None, "Keyword sub-field name", None),
    #         ("keyword_blacklist", None, "Fields that don't have a keyword subfield (wildcards * and ? allowed)", None)
    #         )
    reContainsWildcard = re.compile("(?:(?<!\\\\)|\\\\\\\\)[*?]").search

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.matchKeyword = True
        try:
            self.blacklist = self.keyword_blacklist.split(",")
        except AttributeError:
            self.blacklist = list()

    def containsWildcard(self, value):
        """Determine if value contains wildcard."""
        if type(value) == str:
            res = self.reContainsWildcard(value)
            return res
        else:
            return False


class CarbonBlackQueryBackend(CarbonBlackWildcardHandlingMixin, SingleTextQueryBackend):
    """Converts Sigma rule into CarbonBlack query string. Only searches, no aggregations. Contributed by SOC Prime. https://socprime.com"""
    identifier = "carbonblack"
    active = True

    # reEscape = re.compile("([\s+\\-=!(){}\\[\\]^\"~:/]|(?<!\\\\)\\\\(?![*?\\\\])|\\\\u|&&|\\|\\|)")
    reEscape = re.compile("([\s\s+()\"])")
    reClear = re.compile("[<>]")
    andToken = " AND "
    orToken = " OR "
    notToken = " -"
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = " OR "
    valueExpression = '%s'
    typedValueExpression = {SigmaRegularExpressionModifier: "/%s/"}
    nullExpression = "NOT _exists_:%s"
    notNullExpression = "_exists_:%s"
    mapExpression = "%s:%s"
    mapListsSpecialHandling = False
    escapeCharacters = [
        '\\',
        '"',
        '\'',
        '(',
        ')',
        '[',
        ']',
        '{',
        '}',
        ',',
        '=',
        '<',
        '>',
        '&',
        '|',
        ';',
        ':',
        '-'
    ]

    def __init__(self, *args, **kwargs):
        """Initialize field mappings."""
        super().__init__(*args, **kwargs)
        self.category = None
        self.excluded_fields = None

    def cleanLeading(self, val):
        if val.startswith("*"):
            val = val[1:]
        if val.endswith("*"):
            val = val[:-1]
        return val.strip()

    def escapeCharacter(self, val):
        for ch in self.escapeCharacters:
            val = val.replace(ch, '\\' + ch)
        return val

    def unescapeCharacter(self, val):
        for ch in self.escapeCharacters:
            val = val.replace('\\' + ch, ch)
        return val

    def cleanWhitespace(self, val):
        val = val.replace('*', ' AND ').replace('  ', ' ')
        if re.match('\S+ \S', val):
            matches = re.findall('(?:^|\(| )(.+?)(?:\)| OR| AND|$)', val)
            for strMatch in matches:
                if re.match('\S+ \S', strMatch):
                    strUnescapeMatch = self.unescapeCharacter(strMatch)
                    val = val.replace(strMatch, '"{}"'.format(strUnescapeMatch))
        return val.strip()
    
    def fixWildcards(self, val):
        # prob a better way to do this with SigmaStartswithModifier/SigmaEndswithModifier? idk, fail fast!
        if val.endswith("\\\\"):
            val = val[:-1] + "*"
        if val.startswith("\\\\") and not val.startswith("\\\\\\\\"):
            val = val[2:]
        return val

    def cleanValue(self, val):
        if "[1 to *]" in val:
            self.reEscape = re.compile("([()])")
            val = super().cleanValue(val)
            val = val.strip()
        # else:
        #     self.reEscape = re.compile("([\s\s+()])")
        elif isinstance(val, str):
            val = val.strip()
            val = self.cleanLeading(val)
            val = self.escapeCharacter(val)
            val = self.cleanWhitespace(val)
            val = self.fixWildcards(val)
        return val

    def cleanIPRange(self, value):
        new_value = value
        if isinstance(new_value, str) and value.find('*'):
            sub = value.count('.')
            if value[-2:] == '.*':
                value = value[:-2]
            min_ip = value + '.0' * (4 - sub)
            new_value = min_ip + '/' + str(8 * (4 - sub))
        elif isinstance(new_value, list):
            for index, vl in enumerate(new_value):
                new_value[index] = self.cleanIPRange(vl)

        return new_value

    def generateValueNode(self, node):
        result = self.valueExpression % (str(node))
        if result == "" or result.isspace():
            return '""'
        else:
            if self.matchKeyword:  # don't quote search value on keyword field
                return result
            else:
                return "%s" % result

    def generateMapItemNode(self, node):
        fieldname, value = node
        if fieldname == "EventID" and (type(value) is str or type(value) is int):
            value = str(value)
            fieldname = self.generateEventKey(value)
            value = self.generateEventValue(value)
        if fieldname.lower() in self.excluded_fields:
            return
        else:
            transformed_fieldname = self.fieldNameMapping(fieldname, value)
            if transformed_fieldname == "ipaddr":
                value = self.cleanIPRange(value)
            if (
                self.mapListsSpecialHandling == False
                and type(value) in (str, int, list)
                or self.mapListsSpecialHandling == True
                and type(value) in (str, int)
            ):
                # return self.mapExpression % (transformed_fieldname, self.generateNode(value))
                if isinstance(value, list):
                    return self.generateNode(
                        [
                            self.mapExpression
                            % (transformed_fieldname, self.cleanValue(item))
                            for item in value
                        ]
                    )
                elif isinstance(value, str) or isinstance(value, int):
                    return self.mapExpression % (
                        transformed_fieldname,
                        self.generateNode(self.cleanValue(value)),
                    )
            elif type(value) == list:
                return self.generateMapItemListNode(transformed_fieldname, value)
            elif isinstance(value, SigmaTypeModifier) and not isinstance(
                value, SigmaRegularExpressionModifier
            ):
                return self.generateMapItemTypedNode(transformed_fieldname, value)
            elif value is None:
                return self.nullExpression % (transformed_fieldname,)
            else:
                raise TypeError(
                    "Backend does not support map values of type " + str(type(value))
                )

    def generateNOTNode(self, node):
        expression = super().generateNode(node.item)
        if expression:
            return "(%s%s)" % (self.notToken, expression)

    # Function to upload watchlists through CB API

    def postAPI(self, result, title, desc):
        url = os.getenv("cbapi_watchlist")
        body = {
            "name": title,
            "search_query": "q=" + str(result),
            "description": desc,
            "index_type": "events",
        }
        header = {"X-Auth-Token": os.getenv("APIToken")}
        print(title)
        x = requests.post(url, data=json.dumps(body), headers=header, verify=False)
        print(x.text)

    def generateEventKey(self, value):
        if value in event:
            return event[value][0]
        else:
            return 'eventid'

    def generateEventValue(self, value):
        if value in event:
            return event[value][1]
        else:
            return ''

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        title = sigmaparser.parsedyaml["title"]
        desc = sigmaparser.parsedyaml["description"]
        try:
            self.category = sigmaparser.parsedyaml['logsource'].setdefault(
                'category', None
            )
            self.counted = sigmaparser.parsedyaml.get('counted', None)
            self.excluded_fields = [
                item.lower()
                for item in sigmaparser.config.config.get("excludedfields", [])
            ]
        except KeyError:
            self.category = None
        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed)
            result = ""

            if query is not None:
                result += query
                # self.postAPI(result,title,desc)
            return result
        # if self.category == "process_creation":
        #     for parsed in sigmaparser.condparsed:
        #         query = self.generateQuery(parsed)
        #         result = ""

        #         if query is not None:
        #             result += query
        #         return result
        # else:
        #     raise NotSupportedError("Not supported logsource category.")



# Output backends for sigmac
# Copyright 2016-2018 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Exceptions
class BackendError(Exception):
    """Base exception for backend-specific errors."""
    pass

class NotSupportedError(BackendError):
    """Exception is raised if some output is required that is not supported by the target language."""
    pass

# Exceptions (backend specific Qualys) - TODO: no backend specific exceptions
class PartialMatchError(Exception):
    pass

class FullMatchError(Exception):
    pass    



# Sigma parser
# Copyright 2016-2018 Thomas Patzke, Florian Roth

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sigma.parser.condition import ConditionOR, NodeSubexpression, ConditionNULLValue
from .exceptions import SigmaConfigParseError, FieldMappingError

# Field Mapping Definitions
def FieldMapping(source, target=None):
    """Determines target type and instantiate appropriate mapping type"""
    if target == None:
        return SimpleFieldMapping(source, source)
    elif type(target) == str:
        return SimpleFieldMapping(source, target)
    elif type(target) == list:
        return MultiFieldMapping(source, target)
    elif type(target) == dict:
        return ConditionalFieldMapping(source, target)

class SimpleFieldMapping:
    """1:1 field mapping"""
    target_type = str

    def __init__(self, source, target):
        """Initialization with generic target type check"""
        if type(target) != self.target_type:
            raise TypeError("Target type mismatch: wrong mapping type for this target")
        self.source = source
        self.target = target

    def resolve(self, key, value, sigmaparser):
        """Return mapped field name"""
        return (self.target, value)

    def resolve_fieldname(self, fieldname, sigmaparser=None):
        return self.target

    def __str__(self):  # pragma: no cover
        return "SimpleFieldMapping: {} -> {}".format(self.source, self.target)

class MultiFieldMapping(SimpleFieldMapping):
    """1:n field mapping that expands target field names into OR conditions"""
    target_type = list

    def resolve(self, key, value, sigmaparser):
        """Returns multiple target field names as OR condition"""
        cond = ConditionOR()
        for fieldname in self.target:
            cond.add((fieldname, value))
        return NodeSubexpression(cond)

    def __str__(self):  # pragma: no cover
        return "MultiFieldMapping: {} -> [{}]".format(self.source, ", ".join(self.target))

class ConditionalFieldMapping(SimpleFieldMapping):
    """
    Conditional field mapping:
    * key contains field=value condition, value target mapping
    * key "default" maps when no condition matches
    * if no condition matches and there is no default, don't perform mapping
    """
    target_type = dict

    def __init__(self, source, target):
        """Init table between condition field names and values"""
        super().__init__(source, target)
        self.conditions = dict()    # condition field -> condition value -> target fields
        self.default = None
        for condition, target in self.target.items():
            try:                    # key contains condition (field=value)
                field, value = condition.split("=")
                self.add_condition(field, value, target)
            except ValueError as e:      # no, condition - "default" expected
                if condition == "default":
                    if self.default == None:
                        if type(target) == str:
                            self.default = [ target ]
                        elif type(target) == list:
                            self.default = target
                        else:
                            raise SigmaConfigParseError("Default mapping must be single value or list")
                    else:
                        raise SigmaConfigParseError("Conditional field mapping can have only one default value, use list for multiple target mappings")
                else:
                    raise SigmaConfigParseError("Expected condition or default") from e

    def add_condition(self, field, value, target):
        if field not in self.conditions:
            self.conditions[field] = dict()
        if value not in self.conditions[field]:
            self.conditions[field][value] = list()
        if type(target) == str:
            self.conditions[field][value].append(target)
        elif type(target) == list:
            self.conditions[field][value].extend(target)

    def _targets(self, sigmaparser):
        # build list of matching target mappings
        targets = set()
        for condfield in self.conditions:
            if condfield in sigmaparser.values:
                rulefieldvalues = sigmaparser.values[condfield]
                for condvalue in self.conditions[condfield]:
                    if condvalue in rulefieldvalues:
                        targets.update(self.conditions[condfield][condvalue])
        return targets

    def resolve(self, key, value, sigmaparser):
        targets = self._targets(sigmaparser)
        if len(targets) == 0:       # no matching condition, try with default mapping
            if self.default != None:
                targets = self.default

        if len(targets) == 1:     # result set contains only one target, return mapped item (like SimpleFieldMapping)
            if value is None:
                return ConditionNULLValue(val=list(targets)[0])
            else:
                return (list(targets)[0], value)
        elif len(targets) > 1:        # result set contains multiple targets, return all linked as OR condition (like MultiFieldMapping)
            cond = ConditionOR()
            for target in targets:
                if value is None:
                    cond.add(ConditionNULLValue(val=target))
                else:
                    cond.add((target, value))
            return NodeSubexpression(cond)
        else:                       # no mapping found
            if value is None:
                return ConditionNULLValue(val=key)
            else:
                return (key, value)

    def resolve_fieldname(self, fieldname, sigmaparser=None):
        if sigmaparser is None:
            if self.default != None:
                return self.default
            else:
                return fieldname
        else:
            targets = self._targets(sigmaparser)
            if len(targets) == 0:
                return self.default
            else:
                return targets.pop()       # TODO: this case should be documented

    def __str__(self):  # pragma: no cover
        return "ConditionalFieldMapping: {} -> {}".format(self.source, self.target)

# Field mappimg chain
class FieldMappingChain(object):
    """
    Chain of field mappings and fields used for calculation of a field mapping in chained conversion
    configurations.

    A chain of field mappings may fan out, as one field can map into multiple target fields and these
    must be propagated further. As the whole chain must be completed at configuration parse time, a
    restriction applies to conditional field mappings. These are calculated at rule conversion time and
    therefore it is not possible to decide further mappings after conditionals and these may only appear
    in the last configuration. This case could be solved by calculation of field mappings at rule conversion
    time, but it is not considered as important enough to be implemented at this time.
    """
    def __init__(self, fieldname):
        """Initialize field mapping chain with given field name."""
        self.fieldmappings = set([fieldname])

    def append(self, config):
        """Propagate current possible field mappings with field mapping from configuration"""
        if ConditionalFieldMapping in { type(fieldmapping) for fieldmapping in self.fieldmappings }:   # conditional field mapping appeared before, abort.
            raise FieldMappingError("Conditional field mappings are only allowed in last configuration if configurations are chained.")

        fieldmappings = set()
        if type(self.fieldmappings) == str:
            current_fieldmappings = {self.fieldmappings}
        else:
            current_fieldmappings = self.fieldmappings

        for fieldname in current_fieldmappings:
            mapping = config.get_fieldmapping(fieldname)
            if type(mapping) in (SimpleFieldMapping,  MultiFieldMapping):
                resolved_mapping = mapping.resolve_fieldname(fieldname)
                if type(resolved_mapping) is list:
                    fieldmappings.update(resolved_mapping)
                else:
                    fieldmappings.add(resolved_mapping)
            elif type(mapping) == ConditionalFieldMapping:
                fieldmappings.add(mapping)
            else:
                raise TypeError("Type '{}' is not supported by FieldMappingChain".format(str(type(mapping))))

        if len(fieldmappings) == 1:
            self.fieldmappings = fieldmappings.pop()
        else:
            self.fieldmappings = fieldmappings

    def resolve(self, key, value, sigmaparser):
        if type(self.fieldmappings) == str:     # one field mapping
            return (self.fieldmappings, value)
        elif isinstance(self.fieldmappings, ConditionalFieldMapping):
            logsource = sigmaparser.parsedyaml.get("logsource")
            condition = self.fieldmappings.conditions
            for source_type, logsource_item in logsource.items():
                if condition.get(source_type) and condition.get(source_type, {}).get(logsource_item):
                    new_field = condition.get(source_type, {}).get(logsource_item)
                    self.fieldmappings.default = new_field
            return self.fieldmappings.resolve(self.fieldmappings.source, value, sigmaparser)
        elif isinstance(self.fieldmappings, SimpleFieldMapping):
            return self.fieldmappings.resolve(key, value, sigmaparser)
        elif type(self.fieldmappings) == set:
            cond = ConditionOR()
            for mapping in self.fieldmappings:
                if type(mapping) == str:
                    cond.add((mapping, value))
                elif isinstance(mapping, SimpleFieldMapping):
                    cond.add(mapping.resolve(key, value, sigmaparser))
            return NodeSubexpression(cond)

    def resolve_fieldname(self, fieldname, sigmaparser=None):
        if type(self.fieldmappings) == str:     # one field mapping
            return self.fieldmappings
        elif isinstance(self.fieldmappings, SimpleFieldMapping):
            return self.fieldmappings.resolve_fieldname(fieldname, sigmaparser)
        elif type(self.fieldmappings) == set:
            mappings = set()
            for mapping in self.fieldmappings:
                if type(mapping) == str:
                    mappings.add(mapping)
                elif isinstance(mapping, SimpleFieldMapping):
                    resolved_mapping = mapping.resolve_fieldname(fieldname, sigmaparser)
                    if type(resolved_mapping) is list:
                        mappings.update(resolved_mapping)
                    else:
                        mappings.add(resolved_mapping)
            return list(mappings)

    def __str__(self):  # pragma: no cover
        return "FieldMappingChain: {}".format(self.fieldmappings)



# LimaCharlie backend for sigmac created by LimaCharlie.io
# Copyright 2019 Refraction Point, Inc

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import yaml
from collections import namedtuple
from .base import BaseBackend
from sigma.parser.modifiers.base import SigmaTypeModifier
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier

# A few helper functions for cases where field mapping cannot be done
# as easily one by one, or can be done more efficiently.
def _windowsEventLogArtifactFieldName(fieldName):
    if 'EventID' == fieldName:
        return 'Event/System/EventID'
    return 'Event/EventData/%s' % (fieldName,)

def _windowsEventLogEDRFieldName(fieldName):
    if 'EventID' == fieldName:
        return 'event/EVENT/System/EventID'
    return 'event/EVENT/EventData/%s' % (fieldName,)

def _mapProcessCreationOperations(node):
    # Here we fix some common pitfalls found in rules
    # in a consistent fashion (already processed to D&R rule).

    # First fixup is looking for a specific path prefix
    # based on a specific drive letter. There are many cases
    # where the driver letter can change or where the early
    # boot process refers to it as "\Device\HarddiskVolume1\".
    if ("starts with" == node["op"] and
        "event/FILE_PATH" == node["path"] and
        node["value"].lower().startswith("c:\\")):
        node["op"] = "matches"
        node["re"] = "^(?:(?:.:)|(?:\\\\Device\\\\HarddiskVolume.))\\\\%s" % (re.escape(node["value"][3:]),)
        del(node["value"])

    return node

def _validateSubrules(elem):
    # Make sure that all elements in the rules
    # are actual operators. Otherwise this
    # indicates the Sigma engine provided us
    # with partial contextual information we
    # cannot map to LimaCharlie data with
    # any certainty.
    for sub in elem.get( 'rules', [] ):
        if not isinstance( sub, dict ):
            raise NotImplementedError("Sub-rule does not contain an operator.")

# We support many different log sources so we keep different mapping depending
# on the log source and category.
# The mapping key is product/category/service.
# The mapping value is tuple like:
# - top-level parameters
# - pre-condition is a D&R rule node filtering relevant events.
# - field mappings is a dict with a mapping or a callable to convert the field name.
#       Individual mapping values can also be callabled(fieldname, value) returning a new fieldname and value.
# - isAllStringValues is a bool indicating whether all values should be converted to string.
# - keywordField is the field name to alias for keywords if supported or None if not.
# - postOpMapper is a callback that can modify an operation once it has been generated.
SigmaLCConfig = namedtuple('SigmaLCConfig', [
    'topLevelParams',
    'preConditions',
    'fieldMappings',
    'isAllStringValues',
    'keywordField',
    'postOpMapper',
    'isCaseSensitive',
])
_allFieldMappings = {
    'edr': {
        "windows//": SigmaLCConfig(
            topLevelParams = {
                "event": "WEL",
            },
            preConditions = {
                "op": "is windows",
            },
            fieldMappings = _windowsEventLogEDRFieldName,
            isAllStringValues = True,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = []
        ),
        "windows_defender//": SigmaLCConfig(
            topLevelParams = {
                "event": "WEL",
            },
            preConditions = {
                "op": "is windows",
            },
            fieldMappings = _windowsEventLogEDRFieldName,
            isAllStringValues = True,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = []
        ),
        "windows/process_creation/": SigmaLCConfig(
            topLevelParams = {
                "events": [
                    "NEW_PROCESS",
                    "EXISTING_PROCESS",
                ]
            },
            preConditions = {
                "op": "is windows",
            },
            fieldMappings = {
                "CommandLine": "event/COMMAND_LINE",
                "Image": "event/FILE_PATH",
                "ParentImage": "event/PARENT/FILE_PATH",
                "ParentCommandLine": "event/PARENT/COMMAND_LINE",
                "User": "event/USER_NAME",
                "OriginalFileName": "event/ORIGINAL_FILE_NAME",
                "OriginalFilename": "event/ORIGINAL_FILE_NAME",
                # Custom field names coming from somewhere unknown.
                "NewProcessName": "event/FILE_PATH",
                "ProcessCommandLine": "event/COMMAND_LINE",
                # Another one-off command line.
                "Command": "event/COMMAND_LINE",
            },
            isAllStringValues = False,
            keywordField = "event/COMMAND_LINE",
            postOpMapper = _mapProcessCreationOperations,
            isCaseSensitive = []
        ),
        "windows/dns_query/": SigmaLCConfig(
            topLevelParams = {
                "event": "DNS_REQUEST",
            },
            preConditions = None,
            fieldMappings = {
                "QueryName": "event/DOMAIN_NAME",
                "Image": "event/FILE_PATH",
            },
            isAllStringValues = False,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = []
        ),
        "dns//": SigmaLCConfig(
            topLevelParams = {
                "event": "DNS_REQUEST",
            },
            preConditions = None,
            fieldMappings = {
                "query": "event/DOMAIN_NAME",
            },
            isAllStringValues = False,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = []
        ),
        "linux/process_creation/": SigmaLCConfig(
            topLevelParams = {
                "events": [
                    "NEW_PROCESS",
                    "EXISTING_PROCESS",
                ]
            },
            preConditions = {
                "op": "is linux",
            },
            fieldMappings = {
                "exe": "event/FILE_PATH",
                "type": None,
                "CommandLine": "event/COMMAND_LINE",
                "Image": "event/FILE_PATH",
                "ParentImage": "event/PARENT/FILE_PATH",
                "ParentCommandLine": "event/PARENT/COMMAND_LINE",
                "User": "event/USER_NAME",
                "OriginalFileName": "event/ORIGINAL_FILE_NAME",
                "OriginalFilename": "event/ORIGINAL_FILE_NAME",
            },
            isAllStringValues = False,
            keywordField = 'event/COMMAND_LINE',
            postOpMapper = None,
            isCaseSensitive = ['event/FILE_PATH']
        ),
        "linux/file_event/": SigmaLCConfig(
            topLevelParams = {
                "events": [
                    "FILE_CREATE",
                    "FILE_DELETE",
                    "FILE_MODIFIED",
                    "NEW_DOCUMENT",
                ]
            },
            preConditions = {
                "op": "is linux",
            },
            fieldMappings = {
                "TargetFilename": "event/FILE_PATH",
            },
            isAllStringValues = False,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = ['event/FILE_PATH']
        ),
        "unix//": SigmaLCConfig(
            topLevelParams = {
                "events": [
                    "NEW_PROCESS",
                    "EXISTING_PROCESS",
                ]
            },
            preConditions = {
                "op": "is linux",
            },
            fieldMappings = {
                "exe": "event/FILE_PATH",
                "type": None,
            },
            isAllStringValues = False,
            keywordField = 'event/COMMAND_LINE',
            postOpMapper = None,
            isCaseSensitive = ['event/FILE_PATH']
        ),
        "netflow//": SigmaLCConfig(
            topLevelParams = {
                "event": "NETWORK_CONNECTIONS",
            },
            preConditions = None,
            fieldMappings = {
                "destination.port": "event/NETWORK_ACTIVITY/DESTINATION/PORT",
                "source.port": "event/NETWORK_ACTIVITY/SOURCE/PORT",
            },
            isAllStringValues = False,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = []
        ),
        "/proxy/": SigmaLCConfig(
            topLevelParams = {
                "event": "HTTP_REQUEST",
            },
            preConditions = None,
            fieldMappings = {
                "c-uri|contains": "event/URL",
                "c-uri": "event/URL",
                "URL": "event/URL",
                "cs-uri-query": "event/URL",
                "cs-uri-stem": "event/URL",
            },
            isAllStringValues = False,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = []
        ),
        "macos/process_creation/": SigmaLCConfig(
            topLevelParams = {
                "events": [
                    "NEW_PROCESS",
                    "EXISTING_PROCESS",
                ]
            },
            preConditions = {
                "op": "is mac",
            },
            fieldMappings = {
                "CommandLine": "event/COMMAND_LINE",
                "Commandline": "event/COMMAND_LINE",
                "Image": "event/FILE_PATH",
                "ParentImage": "event/PARENT/FILE_PATH",
                "ParentCommandLine": "event/PARENT/COMMAND_LINE",
                "User": "event/USER_NAME",
                "OriginalFileName": "event/ORIGINAL_FILE_NAME",
                "OriginalFilename": "event/ORIGINAL_FILE_NAME",
                # Custom field names coming from somewhere unknown.
                "NewProcessName": "event/FILE_PATH",
                "ProcessCommandLine": "event/COMMAND_LINE",
                # Another one-off command line.
                "Command": "event/COMMAND_LINE",
            },
            isAllStringValues = False,
            keywordField = "event/COMMAND_LINE",
            postOpMapper = _mapProcessCreationOperations,
            isCaseSensitive = ['event/FILE_PATH']
        ),
        "macos/file_event/": SigmaLCConfig(
            topLevelParams = {
                "events": [
                    "FILE_CREATE",
                    "FILE_DELETE",
                    "FILE_MODIFIED",
                    "NEW_DOCUMENT",
                ]
            },
            preConditions = {
                "op": "is mac",
            },
            fieldMappings = {
                "TargetFilename": "event/FILE_PATH",
            },
            isAllStringValues = False,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = ['event/FILE_PATH']
        ),
    },
    "artifact": {
        "windows//": SigmaLCConfig(
            topLevelParams = {
                "target": "log",
                "log type": "wel",
            },
            preConditions = None,
            fieldMappings = _windowsEventLogArtifactFieldName,
            isAllStringValues = True,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = []
        ),
        "windows_defender//": SigmaLCConfig(
            topLevelParams = {
                "target": "log",
                "log type": "wel",
            },
            preConditions = None,
            fieldMappings = _windowsEventLogArtifactFieldName,
            isAllStringValues = True,
            keywordField = None,
            postOpMapper = None,
            isCaseSensitive = []
        ),
    }
}

class LimaCharlieBackend(BaseBackend):
    """Converts Sigma rule into LimaCharlie D&R rules. Contributed by LimaCharlie. https://limacharlie.io"""
    identifier = "limacharlie"
    active = True
    config_required = False
    default_config = ["limacharlie"]

    options = (
        (
            "lc_target",
            "edr",
            "Generate LimaCharlie D&R rules for the following target, one of: edr, artifact.",
            None,
        ),
    )

    def generate(self, sigmaparser):
        # Take the log source information and figure out which set of mappings to use.
        ruleConfig = sigmaparser.parsedyaml
        ls_rule = ruleConfig['logsource']
        try:
            category = ls_rule['category']
        except KeyError:
            category = ""
        try:
            product = ls_rule['product']
        except KeyError:
            product = ""
        # try:
        #     service = ls_rule['service']
        # except KeyError:
        #     service = ""

        # If there is a timeframe component, we do not currently
        # support it for now.
        if ruleConfig.get( 'detection', {} ).get( 'timeframe', None ) is not None:
            raise NotImplementedError("Timeframes are not supported by backend.")

        # Don't use service for now, most Windows Event Logs
        # uses a different service with no category, since we
        # treat all Windows Event Logs together we can ignore
        # the service.
        service = ""

        # See if we have a definition for the source combination.
        mappingKey = "%s/%s/%s" % (product, category, service)
        topFilter, preCond, mappings, isAllStringValues, keywordField, postOpMapper, isCaseSensitive = _allFieldMappings.get(self.lc_target, {}).get(mappingKey, tuple([None, None, None, None, None, None, None]))
        if mappings is None:
            raise NotImplementedError("Log source %s/%s/%s not supported by backend." % (product, category, service))

        # Field name conversions.
        self._fieldMappingInEffect = mappings

        # LC event type pre-selector for the type of data.
        self._preCondition = preCond

        # Are all the values treated as strings?
        self._isAllStringValues = isAllStringValues

        # Are we supporting keywords full text search?
        self._keywordField = keywordField

        # Call to fixup all operations after the fact.
        self._postOpMapper = postOpMapper

        # Event paths that are case sensitive.
        self._isCaseSensitiveFS = isCaseSensitive

        # Call the original generation code.
        detectComponent = super().generate(sigmaparser)

        # We expect a string (yaml) as output, so if
        # we get anything else we assume it's a core
        # library value and just return it as-is.
        if not isinstance( detectComponent, str):
            return detectComponent

        # This redundant to deserialize it right after
        # generating the yaml, but we try to use the parent
        # official class code as much as possible for future
        # compatibility.
        detectComponent = yaml.safe_load(detectComponent)

        # Check that we got a proper node and not just a string
        # which we don't really know what to do with.
        if not isinstance(detectComponent, dict):
            raise NotImplementedError("Selection combination not supported.")

        # Apply top level filter.
        detectComponent.update(topFilter)

        # Now prepare the Response component.
        respondComponents = [{
            "action": "report",
            "name": ruleConfig["title"],
        }]

        # Add a lot of the metadata available to the report.
        if ruleConfig.get("tags", None) is not None:
            respondComponents[0].setdefault("metadata", {})["tags"] = ruleConfig["tags"]

        if ruleConfig.get("description", None) is not None:
            respondComponents[0].setdefault("metadata", {})["description"] = ruleConfig["description"]

        if ruleConfig.get("references", None) is not None:
            respondComponents[0].setdefault("metadata", {})["references"] = ruleConfig["references"]

        if ruleConfig.get("level", None) is not None:
            respondComponents[0].setdefault("metadata", {})["level"] = ruleConfig["level"]

        if ruleConfig.get("author", None) is not None:
            respondComponents[0].setdefault("metadata", {})["author"] = ruleConfig["author"]

        if ruleConfig.get("falsepositives", None) is not None:
            respondComponents[0].setdefault("metadata", {})["falsepositives"] = ruleConfig["falsepositives"]

        # Assemble it all as a single, complete D&R rule.
        return yaml.safe_dump({
            "detect": detectComponent,
            "respond": respondComponents,
        }, default_flow_style = False)

    def generateQuery(self, parsed):
        # We override the generateQuery function because
        # we generate proper JSON structures internally
        # and only convert to string (yaml) once the
        # whole thing is assembled.
        result = self.generateNode(parsed.parsedSearch)

        if self._preCondition is not None:
            result = {
                "op": "and",
                "rules": [
                    self._preCondition,
                    result,
                ]
            }
            _validateSubrules(result)
            if self._postOpMapper is not None:
                result = self._postOpMapper(result)
        return yaml.safe_dump(result)

    def generateANDNode(self, node):
        generated = [ self.generateNode(val) for val in node ]
        filtered = [ g for g in generated if g is not None ]
        if not filtered:
            return None

        # Map any possible keywords.
        filtered = self._mapKeywordVals(filtered)

        if 1 == len(filtered):
            if self._postOpMapper is not None:
                filtered[0] = self._postOpMapper(filtered[0])
            return filtered[0]
        result = {
            "op": "and",
            "rules": filtered,
        }
        _validateSubrules(result)
        if self._postOpMapper is not None:
            result = self._postOpMapper(result)
        return result

    def generateORNode(self, node):
        generated = [self.generateNode(val) for val in node]
        filtered = [g for g in generated if g is not None]
        if not filtered:
            return None

        # Map any possible keywords.
        filtered = self._mapKeywordVals(filtered)

        if 1 == len(filtered):
            if self._postOpMapper is not None:
                filtered[0] = self._postOpMapper(filtered[0])
            return filtered[0]
        result = {
            "op": "or",
            "rules": filtered,
        }
        _validateSubrules(result)
        if self._postOpMapper is not None:
            result = self._postOpMapper(result)
        return result

    def generateNOTNode(self, node):
        generated = self.generateNode(node.item)
        if generated is None:
            return None
        if not isinstance(generated, dict):
            raise NotImplementedError("Not operator not available on non-dict nodes.")
        generated["not"] = not generated.get("not", False)
        return generated

    def generateSubexpressionNode(self, node):
        return self.generateNode(node.items)

    def generateListNode(self, node):
        return [self.generateNode(value) for value in node]

    def generateMapItemNode(self, node):
        fieldname, value = node

        fieldNameAndValCallback = None

        # The mapping can be a dictionary of mapping or a callable
        # to get the correct value.
        if callable(self._fieldMappingInEffect):
            fieldname = self._fieldMappingInEffect(fieldname)
        else:
            try:
                # The mapping can also be a callable that will
                # return a mapped key AND value.
                if callable(self._fieldMappingInEffect[fieldname]):
                    fieldNameAndValCallback = self._fieldMappingInEffect[fieldname]
                else:
                    fieldname = self._fieldMappingInEffect[fieldname]
            except:
                raise NotImplementedError("Field name %s not supported by backend." % (fieldname,))

        # If fieldname returned is None, it's a special case where we
        # ignore the node.
        if fieldname is None:
            return None

        if isinstance(value, (int, str)):
            if fieldNameAndValCallback is not None:
                fieldname, value = fieldNameAndValCallback(fieldname, value)
            op, newVal = self._valuePatternToLcOp(value)
            newOp = {
                "op": op,
                "path": fieldname,
                "case sensitive": fieldname in self._isCaseSensitiveFS,
            }
            if op == "matches":
                newOp["re"] = newVal
            elif op == "exists":
                # Exists has no value.
                newOp.pop( "case sensitive", None )
            else:
                newOp["value"] = newVal
            if self._postOpMapper is not None:
                newOp = self._postOpMapper(newOp)
            return newOp
        elif isinstance(value, list):
            subOps = []
            for v in value:
                if fieldNameAndValCallback is not None:
                    fieldname, v = fieldNameAndValCallback(fieldname, v)
                op, newVal = self._valuePatternToLcOp(v)
                newOp = {
                    "op": op,
                    "path": fieldname,
                    "case sensitive": fieldname in self._isCaseSensitiveFS,
                }
                if op == "matches":
                    newOp["re"] = newVal
                elif op == "exists":
                    # Exists has no value.
                    newOp.pop( "case sensitive", None )
                else:
                    newOp["value"] = newVal
                if self._postOpMapper is not None:
                    newOp = self._postOpMapper(newOp)
                subOps.append(newOp)
            if 1 == len(subOps):
                return subOps[0]
            return {
                "op": "or",
                "rules": subOps
            }
        elif isinstance(value, SigmaTypeModifier):
            if isinstance(value, SigmaRegularExpressionModifier):
                if fieldNameAndValCallback is not None:
                    fieldname, value = fieldNameAndValCallback(fieldname, value)
                result = {
                    "op": "matches",
                    "path": fieldname,
                    "re": str(value),
                }
                if self._postOpMapper is not None:
                    result = self._postOpMapper(result)
                return result
            else:
                raise TypeError("Backend does not support TypeModifier: %s" % (str(type(value))))
        elif value is None:
            if fieldNameAndValCallback is not None:
                fieldname, value = fieldNameAndValCallback(fieldname, value)
            result = {
                "op": "exists",
                "not": True,
                "path": fieldname,
            }
            if self._postOpMapper is not None:
                result = self._postOpMapper(result)
            return result
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateValueNode(self, node):
        return node

    def _valuePatternToLcOp(self, val):
        # Here we convert the string values supported by Sigma that
        # can include wildcards into either proper values (string or int)
        # or into altered values to be functionally equivalent using
        # a few different LC D&R rule operators.

        # No point evaluating non-strings.
        if not isinstance(val, str):
            return ("is", str(val) if self._isAllStringValues else val)

        # Is there any wildcard in this string? If not, we can short circuit.
        if "*" not in val and "?" not in val:
            return ("is", val)

        if val == "*":
            return ("exists", None)

        # Now we do a small optimization for the shortcut operators
        # available in LC. We try to see if the wildcards are around
        # the main value, but NOT within. If that's the case we can
        # use the "starts with", "ends with" or "contains" operators.
        isStartsWithWildcard = False
        isEndsWithWildcard = False
        tmpVal = val
        if tmpVal.startswith("*"):
            isStartsWithWildcard = True
            tmpVal = tmpVal[1:]
        if tmpVal.endswith("*") and not (tmpVal.endswith("\\*") and not tmpVal.endswith("\\\\*")):
            isEndsWithWildcard = True
            if tmpVal.endswith("\\\\*"):
                # An extra \ had to be there so it didn't escapte the
                # *, but since we plan on removing the *, we can also
                # remove one \.
                tmpVal = tmpVal[:-2]
            else:
                tmpVal = tmpVal[:-1]

        # Check to see if there are any other wildcards. If there are
        # we cannot use our shortcuts.
        if "*" not in tmpVal and "?" not in tmpVal:
            if isStartsWithWildcard and isEndsWithWildcard:
                return ("contains", tmpVal)

            if isStartsWithWildcard:
                return ("ends with", tmpVal)

            if isEndsWithWildcard:
                return ("starts with", tmpVal)

        # This is messy, but it is accurate in generating a RE based on
        # the simplified wildcard system, while also supporting the
        # escaping of those wildcards.
        segments = []
        tmpVal = val
        while True:
            nEscapes = 0
            for i in range(len(tmpVal)):
                # We keep a running count of backslash escape
                # characters we see so that if we meet a wildcard
                # we can tell whether the wildcard is escaped
                # (with odd number of escapes) or if it's just a
                # backslash literal before a wildcard (even number).
                if "\\" == tmpVal[i]:
                    nEscapes += 1
                    continue

                if "*" == tmpVal[i]:
                    if 0 == nEscapes:
                        segments.append(re.escape(tmpVal[:i]))
                        segments.append(".*")
                    elif nEscapes % 2 == 0:
                        segments.append(re.escape(tmpVal[:i - nEscapes]))
                        segments.append(tmpVal[i - nEscapes:i])
                        segments.append(".*")
                    else:
                        segments.append(re.escape(tmpVal[:i - nEscapes]))
                        segments.append(tmpVal[i - nEscapes:i + 1])
                    tmpVal = tmpVal[i + 1:]
                    break

                if "?" == tmpVal[i]:
                    if 0 == nEscapes:
                        segments.append(re.escape(tmpVal[:i]))
                        segments.append(".")
                    elif nEscapes % 2 == 0:
                        segments.append(re.escape(tmpVal[:i - nEscapes]))
                        segments.append(tmpVal[i - nEscapes:i])
                        segments.append(".")
                    else:
                        segments.append(re.escape(tmpVal[:i - nEscapes]))
                        segments.append(tmpVal[i - nEscapes:i + 1])
                    tmpVal = tmpVal[i + 1:]
                    break

                nEscapes = 0
            else:
                segments.append(re.escape(tmpVal))
                break

        val = ''.join(segments)

        return ("matches", val)

    def _mapKeywordVals(self, values):
        # This function ensures that the list of values passed
        # are proper D&R operations, if they are strings it indicates
        # they were requested as keyword matches. We only support
        # keyword matches when specified in the config. We generally just
        # map them to the most common field in LC that makes sense.
        mapped = []

        for val in values:
            # Non-keywords are just passed through.
            if not isinstance(val, str):
                mapped.append(val)
                continue

            if self._keywordField is None:
                raise NotImplementedError("Full-text keyboard searches not supported.")

            # This seems to be indicative only of "keywords" which are mostly
            # representative of full-text searches. We don't support that but
            # in some data sources we can alias them to an actual field.
            op, newVal = self._valuePatternToLcOp(val)
            newOp = {
                "op": op,
                "path": self._keywordField,
            }
            if op == "matches":
                newOp["re"] = newVal
            elif op == "exists":
                # Exists has no value.
                pass
            else:
                newOp["value"] = newVal
            mapped.append(newOp)

        return mapped



import re
import requests
import json
import os
from sigma.config.eventdict import event
from fnmatch import fnmatch

from sigma.parser.rule import SigmaParser
from sigma.backends.base import SingleTextQueryBackend
from sigma.backends.exceptions import NotSupportedError
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from sigma.parser.condition import ConditionOR, ConditionAND, NodeSubexpression

from sigma.parser.modifiers.base import SigmaTypeModifier


class StreamAlertQueryBackend(SingleTextQueryBackend):
    """Converts Sigma rule into StreamAlert code. Not support aggregations. Contributed by AlertIQ."""

    identifier = "streamalert"
    active = True

    andToken = " and "
    orToken = " or "
    notToken = "not "
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = " or "

    valueExpression = '%s'
    valueRawExpression = "record['_raw'].contains(%s)"
    typedValueExpression = {SigmaRegularExpressionModifier: "'%s'"}

    nullExpression = "record.get('%s', None) == None"
    notNullExpression = "record.get('%s', None) != None"
    mapExpression = "record['%s'] == %s"
    endswithExpression = "record['%s'].endswith(%s)"
    startswithExpression = "record['%s'].startswith(%s)"
    containsExpression = "record['%s'].contains(%s)"
    regexExpression = "re.match(%s, record['%s'])"

    mapListsSpecialHandling = False
    escapeCharacters = ['\\', '\'', '\"']
    removeLeadingCharacters = ['*']

    def __init__(self, *args, **kwargs):
        """Initialize field mappings."""
        super().__init__(*args, **kwargs)
        self.excluded_fields = None
        self.is_upper = False
        self.import_libs = ''

    def cleanLeading(self, val: str):
        if val != '*' and any(val.startswith(a) for a in self.removeLeadingCharacters):
            val = val[1:]
        if val != '*' and any(val.endswith(a) for a in self.removeLeadingCharacters):
            val = val[:-1]

        val = val.strip()

        # val = re.sub(r'^\\+', '', val)
        val = re.sub(r'\\+$', '', val)

        if not val.startswith('"'):
            val = '"%s' % re.sub(r'(?<=[^\\])"', '\\"', val)
        if not val.endswith('"') or val == '"' or val.endswith('\\"'):
            val = '%s"' % re.sub(r'(?<=[^\\])"', '\\"', val)

        return val

    def escapeCharacter(self, val):
        for ch in self.escapeCharacters:
            val = val.replace(ch, '\\' + ch)
        return val

    def unescapeCharacter(self, val):
        for ch in self.escapeCharacters:
            val = val.replace('\\' + ch, ch)
        return val

    def cleanWhitespace(self, val):
        val = val.replace('  ', ' ')
        if re.match('\S+ \S', val):
            matches = re.findall('(?:^|\(| )(.+?)(?:\)| or| and|$)', val)
            for strMatch in matches:
                if re.match('\S+ \S', strMatch):
                    strUnescapeMatch = self.unescapeCharacter(strMatch)
                    val = val.replace(strMatch, '{}'.format(strUnescapeMatch))
        return val.strip()

    def cleanValue(self, val):
        if isinstance(val, str):
            val = val.strip()
            val = self.escapeCharacter(val)
            val = self.cleanLeading(val)
            # val = self.cleanWhitespace(val)
        return val

    def cleanIPRange(self, value):
        new_value = value
        if isinstance(new_value, str) and value.find('*'):
            sub = value.count('.')
            if value[-2:] == '.*':
                value = value[:-2]
            min_ip = value + '.0' * (4 - sub)
            new_value = min_ip + '/' + str(8 * (4 - sub))
        elif isinstance(new_value, list):
            for index, vl in enumerate(new_value):
                new_value[index] = self.cleanIPRange(vl)

        return new_value

    def fieldNameMapping(self, fieldname: str, value):
        if fieldname.isupper():
            return fieldname
        new_fieldname = ''
        if self.is_upper:
            for i, a in enumerate(fieldname):
                if a.lower() == 'p' and fieldname[i - 1].lower() == 'i' and i > 0:
                    new_fieldname += a.upper()
                elif a.isupper() and i > 0:
                    new_fieldname += f'_{a}'
                else:
                    new_fieldname += a.upper()

        return new_fieldname if new_fieldname else fieldname

    def mapNode(self, key, value):
        if isinstance(value, str):
            if value != '*' and value.startswith('*') and value.endswith('*'):
                return self.containsExpression % (key, self.cleanValue(value))
            if value.startswith('*') and not value.endswith('*'):
                return self.endswithExpression % (key, self.cleanValue(value))
            if not value.startswith('*') and value.endswith('*'):
                return self.startswithExpression % (key, self.cleanValue(value))

        if isinstance(value, SigmaTypeModifier):
            if isinstance(value, SigmaRegularExpressionModifier):
                self.import_libs += (
                    'import re\n' if 'import re\n' not in self.import_libs else ''
                )
                return self.regexExpression % (self.generateTypedValueNode(value), key)

        return self.mapExpression % (key, self.cleanValue(value))

    def generateValueNode(self, node):
        return (
            self.valueExpression % (str(node))
            if 'record[\'' in str(node)
            else self.valueRawExpression % self.cleanValue(str(node))
        )

    def generateMapItemTypedNode(self, fieldname, value):
        return self.mapExpression % (fieldname, self.generateTypedValueNode(value))

    def generateMapItemNode(self, node):
        fieldname, value = node
        if fieldname.lower() in self.excluded_fields:
            return
        else:
            transformed_fieldname = self.fieldNameMapping(fieldname, value)
            if any(a in transformed_fieldname for a in ["_IP" or "IP_"]):
                value = self.cleanIPRange(value)
            if (
                self.mapListsSpecialHandling == False
                and type(value) in (str, int, list)
            ) or (self.mapListsSpecialHandling == True and type(value) in (str, int)):
                if isinstance(value, list):
                    return self.generateNode(
                        [self.mapNode(transformed_fieldname, a) for a in value]
                    )
                elif isinstance(value, str) or isinstance(value, int):
                    return self.mapNode(transformed_fieldname, value)
            elif type(value) == list:
                return self.generateMapItemListNode(transformed_fieldname, value)
            elif isinstance(value, SigmaTypeModifier):
                return self.mapNode(transformed_fieldname, value)
            elif value is None:
                return self.nullExpression % (transformed_fieldname,)
            else:
                raise TypeError(
                    "Backend does not support map values of type " + str(type(value))
                )

    def generate(self, sigmaparser: SigmaParser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        title = sigmaparser.parsedyaml["title"]

        title_temp = re.sub('\W *', ' ', title).replace('  ', ' ')
        function_name = '_'.join(a.lower() for a in title_temp.split(' '))

        description = sigmaparser.parsedyaml["description"]

        try:
            self.excluded_fields = [
                item.lower()
                for item in sigmaparser.config.config.get("excludedfields", [])
            ]
            service = sigmaparser.parsedyaml['logsource'].get('service', '{service}')
            logsource = ' - '.join(
                '{}:{}'.format(k, v)
                for k, v in sigmaparser.parsedyaml['logsource'].items()
            )
            self.is_upper = 'upper' in sigmaparser.config.config.get("tags", [])
            outputs = sigmaparser.config.config.get("outputs", [])
            publishers = '[{}]'.format(
                ', '.join(sigmaparser.config.config.get("publishers", []))
            )

            self.import_libs = ''
            for publisher in sigmaparser.config.config.get("publishers", []):
                self.import_libs += 'from publishers.general.{0} import {0}\n'.format(
                    publisher
                )
        except KeyError:
            logsource = '{logsource}'
            service = '{service}'

        results = ''
        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed)
            result = '''{import_libs}
from streamalert.shared.rule import rule

"""
Title: {title}
Description: {description}
Log Source: {logsource}
"""
@rule(
    logs=['{service}:events'],
    merge_by_keys=['ALERT_NAME'],
    merge_window_mins=5,
    outputs={outputs},
    publishers={publishers}
)
def {function_name}(record):
    return {query}
'''

            if query is not None:
                results += result.format(
                    import_libs=self.import_libs,
                    logsource=logsource,
                    service=service,
                    query=query,
                    title=title,
                    function_name=function_name,
                    description=description,
                    outputs=outputs,
                    publishers=publishers,
                )
        return results



#!/usr/bin/env python3

import argparse
import glob
import json
import os
import sys

import yaml

level_eq = {
    "informational" : 1,
    "low"           : 2,
    "medium"        : 3,
    "high"          : 4,
    "critical"      : 5
    }

status_eq = {
    "unsupported"   : 1,
    "deprecated"    : 2,
    "experimental"  : 3,
    "test"          : 4,
    "stable"        : 5
        }

def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--rules-directory", "-d", dest="rules_dir", default="rules", help="Directory to read rules from")
    parser.add_argument("--out-file", "-o", dest="out_file", default="heatmap.json", help="File to write the JSON layer to")
    parser.add_argument("--no-comment", dest="no_comment", action="store_true", help="Don't store rule names in comments")
    parser.add_argument("--status-start", "-s",dest="status_start", default="unsupported", help="Check rule with minimun status")
    parser.add_argument("--status-end", "-se",dest="status_end", default="stable", help="Check rule with maximun status")
    parser.add_argument("--level-score", "-l",dest="level_score", action="store_true", help="Score depand form rule level")


    args = parser.parse_args()
    print(args.level_score)
    status_start = status_eq[args.status_start]
    status_end = status_eq[args.status_end]

    rule_files = glob.glob(os.path.join(args.rules_dir, "**/*.yml"), recursive=True)
    techniques_to_rules = {}
    score_to_rules = {}
    curr_max_technique_count = 0
    num_rules_used = 0
    num_rules_no_tags = 0
    num_rules_no_techniques = 0
    for rule_file in rule_files:
        with open(rule_file,encoding='utf-8') as f:
            docs = yaml.load_all(f, Loader=yaml.FullLoader)
            double = False
            for rule in docs:
                if "tags" not in rule :
                    if double == False : # Only 1 warning
                        sys.stderr.write(f"Ignoring rule {rule_file} (no tags)\n")
                        num_rules_no_tags += 1
                    double = True # action globle no tag
                    continue
                if not "status" in rule:
                    status_name = "experimental"
                else:
                    status_name = rule["status"]
                status_nb = status_eq[status_name]
                if status_nb <status_start or status_nb>status_end:
                    sys.stderr.write(f"Ignoring rule {rule_file} filter status : {status_name}\n")
                    continue
                tags = rule["tags"]
                level = rule["level"]
                double = True
                t_tags = False
                for tag in tags:
                    if tag.lower().startswith("attack.t"):
                        t_tags = True
                        technique_id = tag[len("attack."):].upper()
                        num_rules_used += 1
                        if technique_id not in techniques_to_rules:
                            techniques_to_rules[technique_id] = []
                            score_to_rules[technique_id] = []
                        techniques_to_rules[technique_id].append(os.path.basename(rule_file))
                        score_to_rules[technique_id].append(level_eq[level])
                        if args.level_score == True:
                            curr_max_technique_count = max(curr_max_technique_count, sum(score_to_rules[technique_id]))
                        else:
                            curr_max_technique_count = max(curr_max_technique_count, len(techniques_to_rules[technique_id]))
                if t_tags == False:
                    sys.stderr.write(f"Ignoring rule {rule_file} no Techniques in {tags} \n")
                    num_rules_no_techniques += 1

    scores = []
    for technique in techniques_to_rules:
        if args.level_score == True:
            technique_score = sum(score_to_rules[technique])
        else:
            technique_score = len(techniques_to_rules[technique])
        entry = {
            "techniqueID": technique, 
            "score": technique_score, 
        }
        if not args.no_comment:
            entry["comment"] = "\n".join(techniques_to_rules[technique])

        scores.append(entry)

    output = {
        "name": "Sigma rules heatmap",
        "versions": {
            "attack": "10",
            "navigator": "4.4.4",
            "layer": "4.2"
        },
        "domain": "enterprise-attack",
        "description": "Sigma rules heatmap",
        "gradient": {
		    "colors": [
			    "#66b1ffff",
			    "#ff66f4ff",
			    "#ff6666ff"
		    ],
            "maxValue": curr_max_technique_count,
            "minValue": 0
        },

        "techniques": scores,
    }

    with open(args.out_file, "w",encoding="UTF-8") as f:
        f.write(json.dumps(output, indent=4, ensure_ascii=False))
        print(f"[*] Layer file written in {args.out_file} ({str(num_rules_used)} rules)")
        if num_rules_no_tags>0 :
            print(f"[-] Ignored  {num_rules_no_tags} rules without tags")
        else:
            print(f"[*] No rule without tags")
        if num_rules_no_techniques>0:
            print(f"[-] Ignored {num_rules_no_techniques} rules whitout Mitre Technique")
        else:
            print(f"[*] No rule whitout Mitre Technique")

if __name__ == "__main__":
    main()



# Output backends for sigmac
# Copyright 2018 SOC Prime

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
import sigma
from sigma.parser.modifiers.base import SigmaTypeModifier
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier
from .base import SingleTextQueryBackend
from .mixins import MultiRuleOutputMixin


class QRadarBackend(SingleTextQueryBackend):
    """Converts Sigma rule into Qradar saved search. Contributed by SOC Prime. https://socprime.com"""
    identifier = "qradar"
    active = True
    config_required = False
    default_config = ["sysmon", "qradar"]
    reEscape = re.compile('(")')
    reClear = None
    andToken = " and "
    orToken = " or "
    notToken = "not "
    subExpression = "(%s)"
    listExpression = "%s"
    listSeparator = " "
    valueExpression = "\'%s\'"
    keyExpression = "%s"
    nullExpression = "%s is null"
    notNullExpression = "not (%s is null)"
    mapExpression = "%s=%s"
    mapListsSpecialHandling = True
    aql_database = "events"

    def cleanKey(self, key):
        if key == None:
            return ""
        if " " in key:
            key = "\"%s\"" % (key)
            return key
        else:
            return key

    def cleanValue(self, value):
        """Remove quotes in text"""
        return value.replace("\'","\\\'")

    def generateNode(self, node):
        if type(node) == sigma.parser.condition.ConditionAND:
            return self.generateANDNode(node)
        elif type(node) == sigma.parser.condition.ConditionOR:
            return self.generateORNode(node)
        elif type(node) == sigma.parser.condition.ConditionNOT:
            return self.generateNOTNode(node)
        elif type(node) == sigma.parser.condition.ConditionNULLValue:
            return self.generateNULLValueNode(node)
        elif type(node) == sigma.parser.condition.ConditionNotNULLValue:
            return self.generateNotNULLValueNode(node)
        elif type(node) == sigma.parser.condition.NodeSubexpression:
            return self.generateSubexpressionNode(node)
        elif type(node) == tuple:
            return self.generateMapItemNode(node)
        elif type(node) in (str, int):
            return self.generateValueNode(node, False)
        elif type(node) == list:
            return self.generateListNode(node)
        else:
            raise TypeError("Node type %s was not expected in Sigma parse tree" % (str(type(node))))


    def generateMapItemNode(self, node):
        key, value = node
        if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            if type(value) == str and "*" in value:
                value = value.replace("*", "%")
                return "%s ilike %s" % (self.cleanKey(key), self.generateValueNode(value, True))
            elif type(value) in (str, int):
                return self.mapExpression % (self.cleanKey(key), self.generateValueNode(value, True))
            else:
                return self.mapExpression % (self.cleanKey(key), self.generateNode(value))
        elif type(value) == list:
            return self.generateMapItemListNode(key, value)
        elif isinstance(value, SigmaTypeModifier):
            return self.generateMapItemTypedNode(key, value)
        elif value is None:
            return self.nullExpression % (key, )
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateMapItemListNode(self, key, value):
        itemslist = list()
        for item in value:
            if item is None:
                itemslist.append(self.nullExpression % (key))
            elif type(item) == str and "ip" in key and ("/16" in item or "/24" in item):
                itemslist.append("INCIDR(%s, %s)" % (self.generateValueNode(item, True), self.cleanKey(key)))
            elif type(item) == str and "*" in item:
                item = item.replace("*", "%")
                itemslist.append('%s ilike %s' % (self.cleanKey(key), self.generateValueNode(item, True)))
            else:
                itemslist.append('%s = %s' % (self.cleanKey(key), self.generateValueNode(item, True)))
        return '('+" or ".join(itemslist)+')'

    def generateMapItemTypedNode(self, fieldname, value):
        if type(value) == SigmaRegularExpressionModifier:
            regex = str(value)
            # Regular Expressions have to match the full value in QRadar
            if not (regex.startswith('^') or regex.startswith('.*')):
                regex = '.*' + regex
            if not (regex.endswith('$') or regex.endswith('.*')):
                regex = regex + '.*'
            return "%s imatches %s" % (self.cleanKey(fieldname), self.generateValueNode(regex, True))
        else:
            raise NotImplementedError("Type modifier '{}' is not supported by backend".format(value.identifier))

    def generateValueNode(self, node, keypresent):
        if keypresent == False:
            return "UTF8(payload) ilike \'{0}{1}{2}\'".format("%", self.cleanValue(str(node)), "%")
        else:
            return self.valueExpression % (self.cleanValue(str(node)))

    def generateNULLValueNode(self, node):
        return self.nullExpression % (node.item)

    def generateNotNULLValueNode(self, node):
        return self.notNullExpression % (node.item)

    def generateAggregation(self, agg, timeframe='00'):
        if agg == None:
            return ""
        if agg.aggfunc == sigma.parser.condition.SigmaAggregationParser.AGGFUNC_NEAR:
            raise NotImplementedError("The 'near' aggregation operator is not yet implemented for this backend")
        if agg.groupfield == None:
            self.qradarPrefixAgg = "SELECT %s(%s) as agg_val from %s where" % (agg.aggfunc_notrans, self.cleanKey(agg.aggfield), self.aql_database)
            self.qradarSuffixAgg = " group by %s having agg_val %s %s" % (self.cleanKey(agg.aggfield), agg.cond_op, agg.condition)
            return self.qradarPrefixAgg, self.qradarSuffixAgg
        elif agg.groupfield != None and timeframe == '00':
                self.qradarPrefixAgg = " SELECT %s(%s) as agg_val from %s where " % (agg.aggfunc_notrans, self.cleanKey(agg.aggfield), self.aql_database)
                self.qradarSuffixAgg = " group by %s having agg_val %s %s" % (self.cleanKey(agg.groupfield), agg.cond_op, agg.condition)
                return self.qradarPrefixAgg, self.qradarSuffixAgg
        elif agg.groupfield != None and timeframe != None:
            for key, duration in self.generateTimeframe(timeframe).items():
                self.qradarPrefixAgg = " SELECT %s(%s) as agg_val from %s where " % (agg.aggfunc_notrans, self.cleanKey(agg.aggfield), self.aql_database)
                self.qradarSuffixAgg = " group by %s having agg_val %s %s LAST %s %s" % (self.cleanKey(agg.groupfield), agg.cond_op, agg.condition, duration, key)
                return self.qradarPrefixAgg, self.qradarSuffixAgg
        else:
            self.qradarPrefixAgg = " SELECT %s(%s) as agg_val from %s where " % (agg.aggfunc_notrans, self.cleanKey(agg.aggfield), self.aql_database)
            self.qradarSuffixAgg = " group by %s having agg_val %s %s" % (self.cleanKey(agg.groupfield), agg.cond_op, agg.condition)
            return self.qradarPrefixAgg, self.qradarSuffixAgg

    def generateTimeframe(self, timeframe):
        time_unit = timeframe[-1:]
        duration = timeframe[:-1]
        timeframe_object = {}
        if time_unit == "s":
            timeframe_object['seconds'] = int(duration)
        elif time_unit == "m":
            timeframe_object['minutes'] = int(duration)
        elif time_unit == "h":
            timeframe_object['hours'] = int(duration)
        elif time_unit == "d":
            timeframe_object['days'] = int(duration)
        else:
            timeframe_object['months'] = int(duration)
        return timeframe_object

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed, sigmaparser)
            before = self.generateBefore(parsed)
            after = self.generateAfter(parsed)

            result = ""
            if before is not None:
                result = before
            if query is not None:
                result += query
            if after is not None:
                result += after

            return result

    def generateQuery(self, parsed, sigmaparser):
        result = self.generateNode(parsed.parsedSearch)
        self.parsedlogsource = sigmaparser.get_logsource().index
        if any("flow" in i for i in self.parsedlogsource):
            aql_database = "flows"
        else:
            aql_database = "events"

        qradarPrefix="SELECT UTF8(payload) as search_payload"
        try:
            mappedFields = []
            for field in sigmaparser.parsedyaml["fields"]:
                    mapped = sigmaparser.config.get_fieldmapping(field).resolve_fieldname(field, sigmaparser)
                    mappedFields.append(mapped)
                    if " " in mapped and not "(" in mapped:
                        qradarPrefix += ", \"" + mapped + "\""
                    else:
                        qradarPrefix +=  ", " + mapped

        except KeyError:    # no 'fields' attribute
            mapped = None
            pass
        qradarPrefix += " from %s where " % (aql_database)

        try:
            timeframe = sigmaparser.parsedyaml['detection']['timeframe']
        except:
            timeframe = None

        if parsed.parsedAgg and timeframe == None:
            (qradarPrefix, qradarSuffixAgg) = self.generateAggregation(parsed.parsedAgg)
            result = qradarPrefix + result
            result += qradarSuffixAgg
        elif parsed.parsedAgg != None and timeframe != None:
            (qradarPrefix, qradarSuffixAgg) = self.generateAggregation(parsed.parsedAgg, timeframe)
            result = qradarPrefix + result
            result += qradarSuffixAgg
        else:
            result = qradarPrefix + result
        return result



# NetWitness EPL output backend for sigmac
# Copyright 2019 Tarik BOUDJEMAA (@snake-jump)
# Inspired from John Tuckner (@tuckner) NetWitness output backend for sigmac

# NetWitness EPL backend for sigmac uses netwitness-epl.yml config file


# RSA Alerts are generated by Event Processing Language (EPL) , that uses Esper Engine (https://www.espertech.com/esper/)
# For more details see :https://community.rsa.com/docs/DOC-110246  and https://community.rsa.com/docs/DOC-80068

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import re
import sigma
from .base import SingleTextQueryBackend
from .mixins import MultiRuleOutputMixin
from sigma.parser.modifiers.base import SigmaTypeModifier
from sigma.parser.modifiers.type import SigmaRegularExpressionModifier

template="""
module_XXXXX;
@Name('RuleName')
@RSAAlert(oneInSeconds=0) 
SELECT * FROM Event(
    
EXPRESSION
);"""

class NetWitnessEplBackend(SingleTextQueryBackend):
    """Converts Sigma rule into RSA NetWitness EPL . Contributed by @snake-jump"""
    identifier = "netwitness-epl"
    config_required = False
    default_config = ["sysmon","netwitness-epl"]
    active = True
    reEscape = re.compile('(")')
    #reEscape = re.compile("([\\|()\[\]{}.^$+])")
    reClear = None
    andToken = " AND "
    orToken = " OR "
    notToken = "NOT"
    subExpression = "(%s)"
    listExpression = "(%s)"
    listSeparator = ", "
    valueExpression = "\'%s\'"
    keyExpression = "%s"
    nullExpression = "%s is null"
    notNullExpression = "%s is not null"
    mapExpression = "(%s=%s)"
    mapListsSpecialHandling = True

    def generateMapItemNode(self, node):
        key, value = node
        if type(key) != int:
            key = key.replace(".","_")  ## replace . by _ in meta name (RSA EPL)
            key = key.lower()
        if self.mapListsSpecialHandling == False and type(value) in (str, int, list) or self.mapListsSpecialHandling == True and type(value) in (str, int):
            if type(value) == str and "*" in value[1:-1]:
                value = re.sub('([".^$]|\\\\(?![*?]))', '\\\\\g<1>', value)
                value = re.sub('\\*', '.*', value)
                value = re.sub('\\?', '.', value)
                return "(%s REGEXP %s)" %(key, self.generateValueNode(value))
            elif type(value) == str and "*" in value:
                value = re.sub("(\*\\\\)|(\*)", "", value)
                value = self.generateValueNode("%"+value+"%")  # add "%" to construct the like expression  ex: process like %psexesvc%
                return "(%s LIKE %s)" % (key,value)
            elif type(value) in (str, int):
                return self.mapExpression % (key, self.generateValueNode(value))
            else:
                return self.mapExpression % (key, self.generateNode(value))
        elif type(value) == list:
            return self.generateMapItemListNode(key, value)
        elif value is None:
            return self.nullExpression % (key, )
			
        elif type(value) == SigmaRegularExpressionModifier:  ## if value is regex
            regex = str(value)
			## in RSA netwitness EPL regex each backslash must be escaped by backslash
			## ex :  c:\temp\  to regex ->  c:\\temp\\  to RSA EPL regex --> c:\\\\temp\\\\
            regex = regex.replace("\\","\\\\") 
            # Regular Expressions have to match the full value in RSA Netwitness EPL
            if not (regex.startswith('^') or regex.startswith('.*')):
                regex = '.*' + regex
            if not (regex.endswith('$') or regex.endswith('.*')):
                regex = regex + '.*'
            return "(%s REGEXP %s)" %(key, self.generateValueNode(regex))
        else:
            raise TypeError("Backend does not support map values of type " + str(type(value)))

    def generateMapItemListNode(self, key, value):
        equallist = list()
        containlist = list()
        regexlist = list()
        for item in value:
            if type(item) == str and "*" in item[1:-1]:
                item = re.sub('([".^$]|\\\\(?![*?]))', '\\\\\g<1>', item)
                item = re.sub('\\*', '.*', item)
                item = re.sub('\\?', '.', item)
                regexlist.append(self.generateValueNode(item))
            elif type(item) == str and (item.endswith("*") or item.startswith("*")):
                item_temp=item
                item = re.sub("(\*\\\\)|(\*)", "", item)
                if item_temp.endswith("*") and item_temp.startswith("*"):   # pattern begins with "*" and ends with "*"
                    containlist.append(self.generateValueNode('%'+item+'%'))  # add "%" to construct the like expression  ex: process like %psexesvc%
                elif item_temp.startswith("*"):   # pattern don't end with "*"
                    containlist.append(self.generateValueNode('%'+item))
                else: # item_temp.endswith("*")  pattern don't begin with "*"
                    containlist.append(self.generateValueNode(item+'%'))
            else:
                equallist.append(self.generateValueNode(item))
        fmtitems = list()
        if equallist:
            if len(equallist) == 1:
                fmtitems.append("%s = %s" % (key, ", ".join(equallist)))
            else:
                # add "(" and ")" to the first and the last item from the list to have  meta_key IN ('value1','value2')	
                equallist[0]=("("+equallist[0])
                equallist[-1]=(equallist[-1]+")")
                fmtitems.append("%s IN %s" % (key, ", ".join(equallist)))
            
        if containlist:
            fmtitems.append("%s LIKE %s" % (key, (" OR "+key+" LIKE ").join(containlist)))
        if regexlist:
            fmtitems.append("%s REGEXP %s" % (key, "|".join(regexlist)))
        fmtquery = "("+" OR ".join(filter(None, fmtitems))
        # Delete the " ' " from the begin or the end of each regex pattern ex : '.*('patern1'|'patern2').*' --> '.*(patern1|patern2).*'
        fmtquery = re.sub('\'\.\*\(\'','\'.*(',fmtquery)
        fmtquery = re.sub('\'\)\.\*\'',').*\'',fmtquery)
        fmtquery = re.sub('\'\|\'','|',fmtquery)
        fmtquery = fmtquery+')'
        return fmtquery

    def generateValueNode(self, node):
        return self.valueExpression % (str(node))

    def generate(self, sigmaparser):
        """Method is called for each sigma rule and receives the parsed rule (SigmaParser)"""
        for parsed in sigmaparser.condparsed:
            query = self.generateQuery(parsed, sigmaparser)
            query=query.replace('INDEX','`index`')  # index is reserved keyword in Esper and must be escaped
            query=template.replace('EXPRESSION', query)
            try:
                query=query.replace('RuleName', sigmaparser.parsedyaml["title"].replace(" ",""))   # add rule name 
                query=query.replace('module_XXXXX', "module "+sigmaparser.parsedyaml["title"].replace(" ","")) # add rule name
            except:
                print("Error when replacing RuleName by Title from yaml")
                pass
            return query

    def generateQuery(self, parsed, sigmaparser):
        result = self.generateNode(parsed.parsedSearch)
        return result
